#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/file.h>
#include <errno.h>
#include "inno-lib.h"
#include "curl.h"
#include "easy.h"
#include "libxml/parser.h"
#include "libxml/tree.h"
#include <openssl/rc4.h>
#include "process-text-tables.h"

#define A_CHAR                      0x61
#define G_CHAR                      0x67
#define L_CHAR                      0x6C
#define M_CHAR                      0x6D
#define O_CHAR                      0x6F
#define P_CHAR                      0x70
#define Q_CHAR                      0x71
#define S_CHAR                      0x73
#define T_CHAR                      0x74
#define U_CHAR                      0x75
#define SEMI_CHAR                   0x3B
#define AMP_CHAR                    0x26

#define ERROR_SIZE      4096
#define CONFIG_FILE     "/home/lighttpd/server_config/innovaphone.conf"

int trace_level = 0;

const char* errors[] = {
    "",
    "error parsing cdr-xml",
    "error connecting to db",
    "error setting encoding",
    "error preparing transaction for cdr",
    "error preparing transaction for event",
    "error preparing transaction for group",
    "error preparing transaction for cdrp",
    "error preparing transaction for ldap all",
    "error preparing transaction for ldap active",
    "error preparing transaction for cdr guid",
    "error preparing transaction for cdr lock",
    "error DB being restored",
    "error cdr-xml attribute not present",
    "error cdr-xml attribute msg/time missing",
    "error cdr-xml disc/rel event missing",
    "error executing begin cmd",
    "error executing insert cdr",
    "error executing insert event",
    "error executing insert group",
    "error executing insert cdrp",
    "error executing commit cmd",
    "error executing query",
    "document root or script name is empty",
    "no memory for post request",
    "rx incomplete post request",
    "rx empty query",
    "content length field missing",
    "rx unknown request",
    "msg field missing in get request",
    "msg field missing in log request",
    "reporting is not installed",
    "executable not found",
    "no memory for file",
    "no curl instance could be allocated",
    "error executing command",
    "error opening or creating file",
    "unable to open log file",
    "unable to open alarm file",
    "unable to open relay cdr file",
    "error writing to log file",
    "error writing to alarm file",
    "error writing to relay cdr file",
    "error mypbx: parameter not set",
    "error mypbx: action not defined",
    "error mypbx: challenge not created",
    "error mypbx: challenge not valid",
    "error mypbx: cn not matching challenge",
    "error mypbx: auth parameter not set",
    "error mypbx: wrong auth received",
    "error mypbx: fetching missed calls",
    "error mypbx: callback parameters not set",
    "error mypbx: callback update info failed",
    "error mypbx: creating tmp file",
    "error report: no cfg file available",
    "error report: open cfg file",
    "error report: parse cfg file",
    "error report: no memory available",
    "error report: loading images for pdf",
    "error report: calculate dimensions for pdf",
    "error report: open zip file",
    "error report: creating pdf object",
    "error report: creating pdf document",
    "error mypbx: updating clear report time failed",
    "error mypbx: inserting clear report time failed",
    "error executing query lock",
    "error unlicensed cdr",
    "error inserting cdr",
    "error reading LDAP servers",
    "error db handler not valid" /* 66 */
};

int num_errors = sizeof(errors)/sizeof(const char*);

int check_db_restore(void)
{
    char cmdline_link[32];
    char * cmdline;
    int i, tread, read_bytes, result;
    FILE *fp_cmd;
    struct stat status;
    struct timeval tvBegin, tvEnd, tvDiff;
    FILE *fp = fopen("/tmp/config.sh.reporting.postgres", "r");

    if(trace_level & TRACE_TIME) {
        gettimeofday(&tvBegin, NULL);
    }

    if(fp != NULL) {
        char f_line[32];

        if(fgets(f_line, 32, fp) != NULL) {
            int line_len = strlen(f_line);

            f_line[line_len-1] = 0; // overwrite newline character at the end
            sprintf(cmdline_link, "/proc/%d", atoi(f_line));
            result = lstat(cmdline_link, &status);
            if(!result && S_ISDIR(status.st_mode)) {
                sprintf(cmdline_link,"/proc/%d/cmdline", atoi(f_line));
                result = lstat(cmdline_link, &status);
                if(!result && S_ISREG(status.st_mode) != 0) {
                    if((fp_cmd = fopen(cmdline_link, "r")) != NULL) {
                        cmdline = malloc(1024*sizeof(char));
                        read_bytes = 0;
                        do {
                            tread = fread(&cmdline[read_bytes], 1, 128, fp_cmd);
                            read_bytes += tread;
                        } while((tread == 128) && (read_bytes < 800));  // 800 must be enough...
                        for(i=0; i<read_bytes; i++) {
                            if(cmdline[i] == '\0') cmdline[i] = 0x20;
                        }
                        cmdline[read_bytes] = '\0';
                        if(strstr(cmdline,"db-restore") && strstr(cmdline, "innovaphone-reporting")) {
                            fclose(fp_cmd);
                            free(cmdline);
                            if(trace_level & TRACE_TIME) {
                                gettimeofday(&tvEnd, NULL);
                                timeval_subtract(&tvDiff, &tvEnd, &tvBegin);
                                do_log(LOG_REPORTING, "", "procs-found: %ld.%06ld", tvDiff.tv_sec, tvDiff.tv_usec);
                            }
                            return 1; //error(cgi, NULL, "DB restore running", "", query_string, ERROR_DB_BEING_RESTORED, 0, NULL, ERROR_DB_BEING_RESTORED, 1);
                        }
                        fclose(fp_cmd);
                        free(cmdline);
                    }
                }
            }
        }
        fclose(fp);
        remove("/tmp/config.sh.reporting.postgres");
    }
    if(trace_level & TRACE_TIME) {
        gettimeofday(&tvEnd, NULL);
        timeval_subtract(&tvDiff, &tvEnd, &tvBegin);
        do_log(LOG_REPORTING, "", "procs-found: %ld.%06ld", tvDiff.tv_sec, tvDiff.tv_usec);
    }
    return 0;
}

void set_inno_error(inno_config_t * cfg, char * format, ...)
{
    if(cfg->error) free(cfg->error);
    cfg->error = malloc(ERROR_SIZE);

    va_list args;
    va_start (args, format);
    vsnprintf (cfg->error, ERROR_SIZE, format, args);
    va_end (args);
}

void strrev(char * str) 
{
    int len = strlen(str);
    int i;
    char tmp[len + 1];
    char * t = tmp;
    char * s = &str[len - 1];
    for(i = len - 1; i >= 0; i--, t++, s--) {
        *t = *s;
    }
    memcpy(str, tmp, len + 1);
}

int hexstr_to_bytestr(char * hex, unsigned char* byte)
{
    int len = 0;
    int size = 0;
    char ch1, ch2;
    int dig1 = 0, dig2 = 0;
    int i;

    len = strlen(hex);
    if((len & 1) == 1) {       // odd number, musn't occur...
        hex[len] = 0;
        len--;
    }
    size = len>>1;

    for(i=0; i < size; i++, byte++)
    {
        ch1 = *hex;
        hex++;
        ch2 = *hex;
        hex++;
        if(ch1 >= '0' && ch1 <= '9') dig1 = ch1 - '0';
        else if(ch1 >= 'A' && ch1 <= 'F') dig1 = ch1 - 'A' + 10;
        else if(ch1 >= 'a' && ch1 <= 'f') dig1 = ch1 - 'a' + 10;

        if(ch2 >= '0' && ch2 <= '9') dig2 = ch2 - '0';
        else if(ch2 >= 'A' && ch2 <= 'F') dig2 = ch2 - 'A' + 10;
        else if(ch2 >= 'a' && ch2 <= 'f') dig2 = ch2 - 'a' + 10;

        *byte = dig1*16 + dig2;
    }
    return size;
}

char * rc4_decrypt(char * enc_text, char * key)
{
    unsigned char binary[512];
    char pw[512];
    int bin_len = 0;
    int key_len = strlen(key);
    RC4_KEY k;
    char tmp_key[key_len + 1];
    memcpy(tmp_key, key, key_len + 1);
    memset(pw, 0, 512);

    bin_len = hexstr_to_bytestr(enc_text, binary);
    strrev(tmp_key);
    RC4_set_key(&k, key_len, (const unsigned char *)tmp_key);
    RC4(&k, bin_len, binary, (unsigned char *)pw);

    //printf("key: %s bin_len: %d pw: %s\n", key, bin_len, pw);
    //syslog(LOG_INFO, "enc_text: %s key: %s bin_len: %d pw: %s\n", enc_text, key, bin_len, pw);
    char * to = malloc(strlen(pw) + 1);
    strcpy(to, pw);
    return to;
}

char * str_malloc(const char * from)
{
    if(!from) return NULL;
    char * to = malloc(strlen(from) + 1);
    return strcpy(to, from);
}

char * xmlNodeGetContentChar(xmlNodePtr node) {
    char * content = NULL;
    xmlChar * temp = xmlNodeGetContent(node);
    if(temp) {
        content = malloc(xmlStrlen(temp) + 1);
        strcpy(content, (const char *)temp);
        xmlFree(temp);
    }
    else {
        content = str_malloc("");
    }
    return content;
}

int timeval_subtract(struct timeval *result, struct timeval *t2, struct timeval *t1)
{
    long int diff = (t2->tv_usec + 1000000 * t2->tv_sec) - (t1->tv_usec + 1000000 * t1->tv_sec);
    result->tv_sec = diff / 1000000;
    result->tv_usec = diff % 1000000;

    return (diff<0);
}

char * xml_entities(char * inp)
{
    if(!inp) return NULL;

    int i, ind_out;
    int len = strlen(inp);
    char * out = malloc(len * 6 * sizeof(char) + 1);

    out[0] = 0;

    for(i=0,ind_out=0;i<len;i++) {
        switch((int)inp[i]) {
            case 38:    /*&*/
                out[ind_out++] = AMP_CHAR;
                out[ind_out++] = A_CHAR;
                out[ind_out++] = M_CHAR;
                out[ind_out++] = P_CHAR;
                out[ind_out++] = SEMI_CHAR;
                break;
            case 34:    /*"*/
                out[ind_out++] = AMP_CHAR;
                out[ind_out++] = Q_CHAR;
                out[ind_out++] = U_CHAR;
                out[ind_out++] = O_CHAR;
                out[ind_out++] = T_CHAR;
                out[ind_out++] = SEMI_CHAR;
                break;
            case 39:    /*'*/
                out[ind_out++] = AMP_CHAR;
                out[ind_out++] = A_CHAR;
                out[ind_out++] = P_CHAR;
                out[ind_out++] = O_CHAR;
                out[ind_out++] = S_CHAR;
                out[ind_out++] = SEMI_CHAR;
                break;
            case 60:    /*<*/
                out[ind_out++] = AMP_CHAR;
                out[ind_out++] = L_CHAR;
                out[ind_out++] = T_CHAR;
                out[ind_out++] = SEMI_CHAR;
                break;
            case 62:    /*>*/
                out[ind_out++] = AMP_CHAR;
                out[ind_out++] = G_CHAR;
                out[ind_out++] = T_CHAR;
                out[ind_out++] = SEMI_CHAR;
                break;
            default:
                out[ind_out++] = inp[i];
        }
    }
    out[ind_out++] = 0;

    return out;
}

void strcat_xml(char * out, const char * inp) {
    char * str_xml = xml_entities((char *)inp);
    strcat(out, str_xml);
    free(str_xml);
}

void do_log(int log_type, char * application, char * format, ...)
{
    int syslog_type = 0;
    va_list arguments;

    switch(log_type) {
        case LOG_INNOVAPHONE: syslog_type = LOG_LOCAL0; break;
        case LOG_REPORTING: syslog_type = LOG_LOCAL1; break;
        case LOG_ALARMS: syslog_type = LOG_LOCAL2; break;
        case LOG_CDRS: syslog_type = LOG_LOCAL3; break;
        case LOG_LOG: syslog_type = LOG_LOCAL4; break;
    }
    openlog(application, LOG_CONS | LOG_NDELAY | LOG_PID, syslog_type);
    
    va_start ( arguments, format );

    vsyslog(LOG_INFO, format, arguments);

    va_end(arguments);
}

/*  takes the complete query_string and searches for parameter with name 
    returns a malloced copy of the parameter value
*/
char * get_url_param(CURL * conn, char * query_string, char * name, char * from, int log_type)
{
    int name_len = 0;
    char * param = NULL;
    if(name) {
        name_len = strlen(name);
        char * search = malloc(name_len + 2);
        sprintf(search, "%s=", name);
        param = strstr(query_string, search);
        free(search);
    }
    else {
        param = query_string;
    }
    char * ret_param = NULL;
    if(param) {
        char * param_amp = NULL;
        if(name) {
            param += name_len + 1;
            param_amp = strstr(param, "&");
        }
        int param_len = param_amp ? (param_amp - param) : strlen(param);
        if(param_len == 0) {
            return str_malloc("");
        }
        // URL decode the param
        int decode_len;
        char * plus_sign = param;
        while(*plus_sign) {
            if(*plus_sign=='+') *plus_sign=' ';
            plus_sign++;
        }
        char *decoded_param = curl_easy_unescape(conn, param, param_len, &decode_len);
        if (decoded_param){
            ret_param = str_malloc(decoded_param);
            curl_free(decoded_param);
            return ret_param;
        }
        else {
            do_log(log_type, "", "%s: URL decoding failed (param=%s)", from, param);
            return str_malloc(param);   // return still encoded param
        }
    }
    else {
        return NULL;
    }
}

int read_trace_level()
{
    FILE * f = fopen("/var/www/innovaphone/trace.level", "rb");
    if(f) {
        char str_level[11];
        memset(str_level, 0x0, 11*sizeof(char));
        fread(&str_level, 1, sizeof(str_level), f);
        trace_level = atoi(str_level);
        fclose(f);
    }
    return trace_level;
}

int lookup_language(char * lang_str) {
    int i = 0;
    int lang_en = 0;
    for(i = 0; i < NUM_LANGUAGES; i++) {
        if(!strcasecmp(process_language_table[i].lang, "en")) {
            lang_en = i;
            break;
        }
    }
    if(!lang_str) return lang_en;
    for(i = 0; i < NUM_LANGUAGES; i++) {
        if(!strcasecmp(process_language_table[i].lang, lang_str)) {
            return i;
        }
    }
    return lang_en;
}

void cleanup_inno_config(inno_config_t * cfg)
{
    if(cfg->error) free(cfg->error);
    if(cfg->alarm) 
    {
        if(cfg->alarm->ip) free(cfg->alarm->ip);
        if(cfg->alarm->user) free(cfg->alarm->user);
        if(cfg->alarm->password) free(cfg->alarm->password);
        if(cfg->alarm->port) free(cfg->alarm->port);
        free(cfg->alarm);
    }
    free(cfg);
}

inno_config_t * read_inno_config() 
{
    xmlDoc *doc = NULL;
    xmlNodePtr root = NULL;
    xmlNodePtr node = NULL;
    xmlNodePtr child = NULL;
    
    inno_config_t * cfg = malloc(sizeof(inno_config_t));
    memset(cfg, 0, sizeof(inno_config_t));

    if ((doc = xmlReadFile(CONFIG_FILE, NULL, 0)) == NULL) {
        set_inno_error(cfg, "Couldn't read innovaphone config file '%s'.", CONFIG_FILE);
        return cfg;
    } 

    root = xmlDocGetRootElement(doc);
    if(root == NULL) {
        set_inno_error(cfg, "No root element found in innovaphone config file '%s'.", CONFIG_FILE);
        return cfg;
    }

    cfg->alarm = malloc(sizeof(alarm_server_t));
    memset(cfg->alarm, 0, sizeof(alarm_server_t));
    
    for(node = root->children; node != NULL; node = node->next)
    {
        if(!xmlStrcasecmp(node->name, (xmlChar *)"alarm-server")) {
            for(child = node->children; child != NULL; child = child->next) 
            {
                if(!xmlStrcasecmp(child->name, (xmlChar *)"ip")) {
                    cfg->alarm->ip = xmlNodeGetContentChar(child);
                }
                else if(!xmlStrcasecmp(child->name, (xmlChar *)"user")) {
                    cfg->alarm->user = xmlNodeGetContentChar(child);
                }
                else if(!xmlStrcasecmp(child->name, (xmlChar *)"password")) {
                    cfg->alarm->password = xmlNodeGetContentChar(child);
                }
                else if(!xmlStrcasecmp(child->name, (xmlChar *)"port")) {
                    cfg->alarm->port = xmlNodeGetContentChar(child);
                }
                else if(!xmlStrcasecmp(child->name, (xmlChar *)"https")) {
                    char * https = xmlNodeGetContentChar(child);
                    if(!strcasecmp(https, "1") || !strcasecmp(https, "true")) {
                        cfg->alarm->https = 1;
                    }
                    free(https);
                }
            }
        }
    }
    
    xmlFreeDoc(doc);
    doc = NULL;

    if(cfg->alarm->password && *cfg->alarm->password &&\
       cfg->alarm->ip && *cfg->alarm->ip) {
        char * decrypted = rc4_decrypt(cfg->alarm->password, cfg->alarm->ip);
        free(cfg->alarm->password);
        cfg->alarm->password = decrypted;
    }

    return cfg;
}
